from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import Callable, Generic, TypeVar

from .core import Link, List, Variable
from .exc import VariableNotSet

__all__ = ["Var", "SeparatedList", "Ref"]


T = TypeVar("T")


@dataclass(frozen=True)
class Var(Generic[T], Variable[T]):
    name: str
    parser: Callable[[str], T]

    def value(self) -> T:
        try:
            return self.parser(os.environ[self.name])
        except KeyError:
            raise VariableNotSet(self.name, repr(self))

    def __str__(self) -> str:
        return f"env.Var({self.name},{self.parser.__name__})"

    def __repr__(self) -> str:
        return f"env.Var({self.name},{self.parser.__name__})"


def _splitter_factory(sep: str) -> Callable[[str], list[str]]:
    def split_after_strip_if_not_empty(src: str) -> list[str]:
        if src := src.strip():
            return src.split(sep)
        else:
            return []

    return split_after_strip_if_not_empty


@dataclass(frozen=True)
class SeparatedList(List[T, str, str]):
    split_sep: str = ","
    splitter: Callable[[str], list[str]] = _splitter_factory(split_sep)

    def __str__(self) -> str:
        return f"env.SeparatedList({self.src},{self.splitter.__name__},{self.split_sep},{self.parser.__name__})"

    def __repr__(self) -> str:
        return f"env.SeparatedList({repr(self.src)},{self.splitter.__name__},{self.split_sep},{self.parser.__name__})"


@dataclass(frozen=True)
class Ref(Link[T, str]):
    src: Variable[str]
    parser: Callable[[str], T] = field(kw_only=True)
    linker: Callable[[str], Var[T]] = Var.factory(parser=parser)

    def __str__(self) -> str:
        return f"env.Ref({self.src},{self.linker.__name__},{self.parser.__name__})"

    def __repr__(self) -> str:
        return f"env.Ref({repr(self.src)},{self.linker.__name__},{self.parser.__name__})"



from __future__ import annotations

import abc
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import partial, wraps
from typing import Any, Callable, Generic, Self, TypeVar, cast, overload

from .exc import StandbyError, VariableNotSet

P = TypeVar("P")
S = TypeVar("S")
T = TypeVar("T")


class Variable(Generic[T], ABC):
    @abstractmethod
    def value(self) -> T:  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def __str__(self) -> str:  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def __repr__(self) -> str:  # pragma: no cover
        raise NotImplementedError()

    def __call__(self) -> T | None:
        return self.value()

    @overload
    def __get__(self, instance: None, owner: type[object]) -> T: ...

    @overload
    def __get__(self, instance: object, owner: type[object]) -> T: ...

    def __get__(self, instance: object | None, owner: type[object]) -> T | None:
        return self()

    def __rshift__(self, hint: Hint[T]) -> Variable[T]:
        return hint.wrap(self)

    def __or__(self, backup: Variable[T]) -> _BackupWrapper[T]:
        return _BackupWrapper(self, backup)

    def __invert__(self) -> T:
        return cast(T, self)

    @classmethod
    def factory(cls, *args: Any, **kwargs: Any) -> Callable[[Any], Self]:
        return wraps(cls)(partial(cls, *args, **kwargs))


@dataclass(frozen=True)
class _BackupWrapper(Variable[T]):
    variable: Variable[T]
    backup: Variable[T]

    def __str__(self) -> str:
        return repr(self.backup)

    def __repr__(self) -> str:
        return f"{repr(self.variable)}|{self}"

    def value(self) -> T:
        try:
            return self.variable.value()
        except VariableNotSet:
            return self.backup.value()

    def __call__(self) -> T | None:
        try:
            value = self.variable()
        except VariableNotSet:
            value = self.backup()
            if value is None:
                # VariableNotSet means that wrapped variable is required
                raise  # even if backup is not
            return value
        return self.backup() if value is None else value


class Hint(Generic[T], abc.ABC):
    @abc.abstractmethod
    def wrap(self, variable: Variable[T]) -> Variable[T]:  # pragma: no cover
        raise NotImplementedError()


@dataclass(frozen=True)
class Const(Variable[T]):
    val: T

    def value(self) -> T:
        return self.val

    def __str__(self) -> str:
        return f"Const({self.val})"

    def __repr__(self) -> str:
        return f"Const[{type(self.val).__name__}]({self.val})"


@dataclass(frozen=True)
class List(Generic[T, P, S], Variable[list[T]]):
    src: Variable[S]
    splitter: Callable[[S], list[P]] = field(kw_only=True)
    parser: Callable[[P], T] = field(kw_only=True)

    def value(self) -> list[T]:
        try:
            return list(map(self.parser, self.splitter(self.src.value())))
        except StandbyError as exc:
            exc.args += (repr(self),)
            raise exc

    def __call__(self) -> list[T] | None:
        try:
            value = self.src()
        except StandbyError as exc:
            exc.args += (repr(self),)
            raise exc
        return list(map(self.parser, self.splitter(value))) if value is not None else None

    def __str__(self) -> str:
        return f"List({str(self.src)},{self.splitter.__name__},{self.parser.__name__})"

    def __repr__(self) -> str:
        return f"List({repr(self.src)},{self.splitter.__name__},{self.parser.__name__})"


@dataclass(frozen=True)
class Link(Generic[T, S], Variable[T]):
    src: Variable[S]
    linker: Callable[[S], Variable[T]] = field(kw_only=True)

    def value(self) -> T:
        try:
            return self.linker(self.src.value()).value()
        except StandbyError as exc:
            exc.args += (repr(self),)
            raise exc

    def __call__(self) -> T | None:
        try:
            if src := self.src():
                return self.linker(src)()
            return None
        except StandbyError as exc:
            exc.args += (repr(self),)
            raise exc

    def __str__(self) -> str:
        return f"Link({self.src},{self.linker.__name__})"

    def __repr__(self) -> str:
        return f"Link({repr(self.src)},{self.linker.__name__})"



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, TypeVar

from .core import Hint, Variable
from .exc import ValueNotValid, VariableNotSet

__all__ = ["Default", "Required", "Validated"]


R = TypeVar("R")
T = TypeVar("T")


@dataclass(frozen=True)
class Default(Hint[T]):
    value: T

    def wrap(self, variable: Variable[T]) -> Variable[T]:
        return self._Wrapper(variable, self.value)

    @dataclass(frozen=True)
    class _Wrapper(Variable[R]):
        variable: Variable[R]
        default: R

        def value(self) -> R:
            try:
                return self.variable.value()
            except VariableNotSet:
                return self.default

        def __str__(self) -> str:
            return f"Default({self.default})"

        def __repr__(self) -> str:
            return f"{repr(self.variable)}>>{self}"

        def __call__(self) -> R | None:
            try:
                value = self.variable()
            except VariableNotSet:
                return self.default
            return self.default if value is None else value


@dataclass(frozen=True)
class Required(Hint):
    is_required: bool = True

    def wrap(self, variable: Variable[T]) -> Variable[T]:
        return variable if self.is_required else self._Wrapper(variable)

    @dataclass(frozen=True)
    class _Wrapper(Variable[T]):
        variable: Variable[T]

        def __str__(self) -> str:
            return f"Required({False})"

        def __repr__(self) -> str:
            return f"{repr(self.variable)}>>{self}"

        def value(self) -> T:
            return self.variable.value()

        def __call__(self) -> T | None:
            try:
                return self.variable()
            except VariableNotSet:
                return None


@dataclass(frozen=True)
class Validated(Hint[T]):
    validator: Callable[[T], bool]
    raises: bool = True

    def wrap(self, variable: Variable[T]) -> Variable[T]:
        return self._Wrapper(variable, self.validator, self.raises)

    @dataclass(frozen=True)
    class _Wrapper(Variable[R]):
        variable: Variable[R]
        validator: Callable[[R], bool]
        raises: bool = True

        def __str__(self) -> str:
            return (f"Validated({self.validator.__name__},"
                    f"{'raises' if self.raises else 'returns None'} on invalid value)")

        def __repr__(self) -> str:
            return f"{repr(self.variable)}>>{self}"

        def value(self) -> R:
            if self.validator(value := self.variable.value()):
                return value
            else:
                raise ValueNotValid(value)

        def __call__(self) -> R | None:
            value = self.variable()
            if value is None:
                return None
            elif self.validator(value):
                return value
            elif self.raises:
                raise ValueNotValid(value, repr(self))
            else:
                return None
